#!/bin/env python


import io
import logging
import os
import sys
import time

from requests.exceptions import (
    HTTPError,
)
from requests_toolbelt import MultipartEncoder
from .session import APISession


__all__ = [
    'APIClient',
    'ApiEndpoint',
    'API_analyses',
    'API_models',
    'API_portfolios',
    'FileEndpoint',
]

# --- API Endpoint mapping to functions ------------------------------------- #


class ApiEndpoint(object):
    """
    Used to Implement the default requests common to all Oasis API
    End points.
    """
    def __init__(self, session, url_endpoint, logger=None):
        self.logger = logger or logging.getLogger()
        self.session = session
        self.url_endpoint = url_endpoint

    def create(self, data):
        return self.session.post(self.url_endpoint, json=data)

    def get(self, ID=None):
        if ID:
            return self.session.get('{}{}/'.format(self.url_endpoint, ID))
        return self.session.get(self.url_endpoint)

    def delete(self, ID):
        return self.session.delete('{}{}/'.format(self.url_endpoint, ID))


class FileEndpoint(object):
    """
    File Resources Endpoint for Upload / Downloading
    """
    def __init__(self, session, url_endpoint, url_resource, logger=None):
        self.logger = logger or logging.getLogger()

        self.session = session
        self.url_endpoint = url_endpoint
        self.url_resource = url_resource

    def _build_url(self, ID):
        return '{}{}/{}'.format(
            self.url_endpoint,
            ID,
            self.url_resource
        )

    def upload(self, ID, file_path, content_type='text/csv'):
        try:
            abs_fp = os.path.realpath(os.path.expanduser(file_path))
            with io.open(abs_fp, 'rb') as f:
                m = MultipartEncoder(fields={'file': (os.path.basename(file_path), f, content_type)})
                r = self.session.post(self._build_url(ID), data=m, headers={'Content-Type': m.content_type})
                r.raise_for_status()
                return r
        except HTTPError as e:
            err_msg = 'File upload Failed: {}, file: {},  url: {}:'.format(r.status_code, file_path, r.url)
            self.logger.error(r.text)
            self.logger.error(err_msg)

    def download(self, ID, file_path, overwrite=True, chuck_size=1024):
        abs_fp = os.path.realpath(os.path.expanduser(file_path))

        # Check and create base dir
        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path))

        # Check if file exists
        if os.path.exists(abs_fp) and not overwrite:
            if overwrite:
                os.remove(abs_fp)
            else:
                error_message = 'Local file alreday exists: {}'.format(abs_fp)
                raise IOError(error_message)

        with io.open(abs_fp, 'wb') as f:
            r = self.session.get(self._build_url(ID), stream=True)
            for chunk in r.iter_content(chunk_size=chuck_size):
                f.write(chunk)
            return r

    def get(self, ID):
        # fetch file into memory
        return self.session.get(self._build_url(ID))

    def post(self, ID, data_object, content_type='application/json'):
        # Update data as object -
        # https://toolbelt.readthedocs.io/en/latest/uploading-data.html
        m = MultipartEncoder(fields={'file': ('data', data_object, content_type)})
        r = self.session.post(
            self._build_url(ID),
            data=m,
            headers={'Content-Type': m.content_type}
        )
        if not r.ok:
            err_msg = 'Data_Object upload Failed'
            self.logger.error(err_msg)
            # self._logger.error(error_message)
            # raise OasisException(error_message)
        return r

    def delete(self, ID):
        return self.session.delete(self._build_url(ID))


class API_models(ApiEndpoint):
    def __init__(self, session, url_endpoint):
        super(API_models, self).__init__(session, url_endpoint)
        self.resource_file = FileEndpoint(self.session, self.url_endpoint, 'resource_file/')

    def search(self, metadata):
        search_string = None
        for key in metadata:
            if not search_string:
                search_string = '?{}={}'.format(key, metadata[key])
            else:
                search_string += '&{}={}'.format(key, metadata[key])
        return self.session.get('{}{}'.format(self.url_endpoint, search_string))

    def create(self, supplier_id, model_id, version_id):
        data = {"supplier_id": supplier_id,
                "model_id": model_id,
                "version_id": version_id}
        return self.session.post(self.url_endpoint, json=data)

    def update(self, ID, supplier_id, model_id, version_id):
        data = {"supplier_id": supplier_id,
                "model_id": model_id,
                "version_id": version_id}
        return self.session.put('{}{}/'.format(self.url_endpoint, ID), json=data)


class API_portfolios(ApiEndpoint):

    def __init__(self, session, url_endpoint):
        super(API_portfolios, self).__init__(session, url_endpoint)
        self.accounts_file = FileEndpoint(self.session, self.url_endpoint, 'accounts_file/')
        self.location_file = FileEndpoint(self.session, self.url_endpoint, 'location_file/')
        self.reinsurance_info_file = FileEndpoint(self.session, self.url_endpoint, 'reinsurance_info_file/')
        self.reinsurance_source_file = FileEndpoint(self.session, self.url_endpoint, 'reinsurance_source_file/')

    def search(self, metadata):
        search_string = None
        for key in metadata:
            if not search_string:
                search_string = '?{}={}'.format(key, metadata[key])
            else:
                search_string += '&{}={}'.format(key, metadata[key])
        return self.session.get('{}{}'.format(self.url_endpoint, search_string))

    def create(self, name):
        data = {"name": name}
        return self.session.post(self.url_endpoint, json=data)

    def update(self, ID, name):
        data = {"name": name}
        return self.session.put('{}{}/'.format(self.url_endpoint, ID), json=data)

    def create_analyses(self, ID, name, model_id):
        """ Create new analyses from Exisiting portfolio
        """
        data = {"name": name,
                "model": model_id}
        return self.session.post('{}{}/create_analysis/'.format(self.url_endpoint, ID), json=data)


class API_analyses(ApiEndpoint):

    def __init__(self, session, url_endpoint):
        super(API_analyses, self).__init__(session, url_endpoint)
        self.input_errors_file = FileEndpoint(self.session, self.url_endpoint, 'input_errors_file/')
        self.input_file = FileEndpoint(self.session, self.url_endpoint, 'input_file/')
        self.input_generation_traceback_file = FileEndpoint(self.session, self.url_endpoint, 'input_generation_traceback_file/')
        self.output_file = FileEndpoint(self.session, self.url_endpoint, 'output_file/')
        self.run_traceback_file = FileEndpoint(self.session, self.url_endpoint, 'run_traceback_file/')
        self.settings_file = FileEndpoint(self.session, self.url_endpoint, 'settings_file/')

    def search(self, metadata):
        search_string = None
        for key in metadata:
            if not search_string:
                search_string = '?{}={}'.format(key, metadata[key])
            else:
                search_string += '&{}={}'.format(key, metadata[key])
        return self.session.get('{}{}'.format(self.url_endpoint, search_string))

    def create(self, name, portfolio_id, model_id):
        data = {"name": name,
                "portfolio": portfolio_id,
                "model": model_id}
        return self.session.post(self.url_endpoint, json=data)

    def update(self, ID, name, portfolio_id, model_id):
        data = {"name": name,
                "portfolio": portfolio_id,
                "model": model_id}
        return self.session.put('{}{}/'.format(self.url_endpoint, ID), json=data)

    def status(self, ID):
        return self.get(ID).json()['status']

    def generate(self, ID):
        return self.session.post('{}{}/generate_inputs/'.format(self.url_endpoint, ID), json={})

    def generate_cancel(self, ID):
        return self.session.post('{}{}/cancel_generate_inputs/'.format(self.url_endpoint, ID), json={})

    def run(self, ID):
        return self.session.post('{}{}/run/'.format(self.url_endpoint, ID), json={})

    def run_cancel(self, ID):
        return self.session.post('{}{}/cancel/'.format(self.url_endpoint, ID), json={})

# --- API Main Client ------------------------------------------------------- #


class APIClient(object):
    def __init__(self, api_url, api_ver, username, password, timeout=25, logger=None):
        self.logger = logger or logging.getLogger()

        self.api = APISession(api_url, username, password, timeout)
        self.models = API_models(self.api, '{}{}/models/'.format(self.api.url_base, api_ver))
        self.portfolios = API_portfolios(self.api, '{}{}/portfolios/'.format(self.api.url_base, api_ver))
        self.analyses = API_analyses(self.api, '{}{}/analyses/'.format(self.api.url_base, api_ver))

    def upload_inputs(self, portfolio_name=None, portfolio_id=None,
                      location_fp=None, accounts_fp=None, ri_info_fp=None, ri_scope_fp=None):

        if not portfolio_name:
            portfolio_name = time.strftime("Portfolio_%d%m%Y-%H%M%S")

        try:
            if portfolio_id:
                self.logger.info('Updating exisiting portfolio')
                portfolio = self.portfolios.update(portfolio_id, portfolio_name)
            else:
                self.logger.info('Creating portfolio')
                portfolio = self.portfolios.create(portfolio_name)
                portfolio_id = portfolio.json()['id']

            # Check or create portfolio
            if not portfolio.ok:
                err_msg = "Failed to find matching `portfolio_id = {}`".format(portfolio_id)
                self.logger.error(err_msg)
                # raise OasisException()

            # Upload exposure
            if location_fp:
                self.portfolios.location_file.upload(portfolio_id, location_fp)
                self.logger.info("File uploaded: {}".format(location_fp))
            if accounts_fp:
                self.portfolios.accounts_file.upload(portfolio_id, accounts_fp)
                self.logger.info("File uploaded: {}".format(accounts_fp))
            if ri_info_fp:
                self.portfolios.reinsurance_info_file.upload(portfolio_id, ri_info_fp)
                self.logger.info("File uploaded: {}".format(ri_info_fp))
            if ri_scope_fp:
                self.portfolios.reinsurance_source_file.upload(portfolio_id, ri_scope_fp)
                self.logger.info("File uploaded: {}".format(ri_scope_fp))
            return portfolio.json()
        except HTTPError as e:
            self.logger.error(e)
            self.logger.error('upload_inputs: failed')
            sys.exit(1)

    def create_analysis(self, portfolio_id, model_id, analysis_name=None, analysis_settings_fp=None):
        try:
            if not analysis_name:
                analysis_name = time.strftime("Analysis_%d%m%Y-%H%M%S")

            r = self.models.get(model_id)
            r.raise_for_status()

            r = self.portfolios.get(portfolio_id)
            r.raise_for_status()

            r = self.analyses.create(analysis_name, portfolio_id, model_id)
            r.raise_for_status()
            analyses = r.json()

            if analysis_settings_fp:
                r = self.analyses.settings_file.upload(analyses['id'], analysis_settings_fp, 'application/json')
                self.logger.info("File uploaded: {}".format(analysis_settings_fp))

            return analyses
        except HTTPError as e:
            err_msg = 'API Error: {}, url: {}, msg: {}'.format(r.status_code, r.url, r.text)
            self.logger.error(err_msg)
            self.logger.error('create_analysis: failed ')
            sys.exit(1)

    # BLOCKING CALL
    def run_generate(self, analysis_id, poll_interval=5):
        """
        Generates the inputs for the analysis based on the portfolio.
        The analysis must have one of the following statuses, `NEW`, `INPUTS_GENERATION_ERROR`,
        `INPUTS_GENERATION_CANCELED`, `READY`, `RUN_COMPLETED`, `RUN_CANCELLED` or
        `RUN_ERROR`.
        """

        try:
            r = self.analyses.generate(analysis_id)
            r.raise_for_status()
            analysis = r.json()
            self.logger.info('Inputs Generation: Started (id={})'.format(analysis_id))
            while True:
                if analysis['status'] in ['READY']:
                    self.logger.info('Inputs Generation: Complete (id={})'.format(analysis_id))
                    return True

                elif analysis['status'] in ['INPUTS_GENERATION_CANCELED']:
                    self.logger.info('Input Generation: Cancelled (id={})'.format(analysis_id))
                    return False

                elif analysis['status'] in ['INPUTS_GENERATION_ERROR']:
                    self.logger.info('Input Generation: failed (id={})'.format(analysis_id))
                    error_trace = self.analyses.input_generation_traceback_file.get(analysis_id).text
                    self.logger.error(error_trace)
                    return False

                elif analysis['status'] in ['INPUTS_GENERATION_STARTED']:
                    # self.logger.debug('Polling - status: {}'.format(analysis['status']))
                    time.sleep(poll_interval)
                    r = self.analyses.get(analysis_id)
                    r.raise_for_status()
                    analysis = r.json()
                    continue

                else:
                    err_msg = "Inputs Generation: Unknown State'{}'".format(analysis['status'])
                    # Raise oasis Execption
                    # Error -- Raise execption Unknown Analysis  State
        except HTTPError as e:
            err_msg = 'API Error: {}, url: {}, msg: {}'.format(r.status_code, r.url, r.text)
            self.logger.error(err_msg)
            self.logger.error('run_generate: failed')
            sys.exit(1)

    # BLOCKING CALL
    def run_analysis(self, analysis_id, analysis_settings_fp=None, poll_interval=5):
        """
        Runs all the analysis. The analysis must have one of the following
        statuses, `NEW`, `RUN_COMPLETED`, `RUN_CANCELLED` or
        `RUN_ERROR`
        """

        try:
            if analysis_settings_fp:
                r = self.analyses.settings_file.upload(analysis_id, analysis_settings_fp, 'application/json')
                self.logger.info("File uploaded: {}".format(analysis_settings_fp))

            r = self.analyses.run(analysis_id)
            r.raise_for_status()
            analysis = r.json()
            self.logger.info('Analysis Run: Started (id={})'.format(analysis_id))

            while True:
                if analysis['status'] in ['RUN_COMPLETED']:
                    self.logger.info('Analysis Run: Complete (id={})'.format(analysis_id))
                    return True

                elif analysis['status'] in ['RUN_CANCELLED']:
                    self.logger.info('Analysis Run: Cancelled (id={})'.format(analysis_id))
                    return False

                elif analysis['status'] in ['RUN_ERROR']:
                    self.logger.error('Analysis Run: failed (id={})'.format(analysis_id))
                    error_trace = self.analyses.run_traceback_file.get(analysis_id).text
                    self.logger.error(error_trace)
                    return False

                elif analysis['status'] in ['RUN_STARTED']:
                    # self.logger.debug('Polling - status: {}'.format(analysis['status']))
                    time.sleep(poll_interval)
                    r = self.analyses.get(analysis_id)
                    r.raise_for_status()
                    analysis = r.json()
                    continue

                else:
                    err_msg = "Execution status in Unknown State: '{}'".format(analysis['status'])
                    # Raise oasis Execption
                    # Error -- Raise execption Unknown Analysis  State
        except HTTPError as e:
            err_msg = 'API Error: {}, url: {}, msg: {}'.format(r.status_code, r.url, r.text)
            self.logger.error(err_msg)
            self.logger.error('run_analysis: failed')
            sys.exit(1)

    def download_output(self, analysis_id, download_path, filename=None, clean_up=False, overwrite=True):
        if not filename:
            filename = 'analysis_{}_output'.format(analysis_id)
        try:
            output_file = os.path.join(download_path, filename + '.tar')
            r = self.analyses.output_file.download(ID=analysis_id, file_path=output_file, overwrite=overwrite)
            r.raise_for_status()
            self.logger.info('Analysis Download output: filename={}, (id={}'.format(output_file, analysis_id))
            if clean_up:
                r = self.analyses.delete(analysis_id)
                r = self.analyses.output_file.delete(analysis_id)
                r = self.analyses.input_file.delete(analysis_id)
        except HTTPError as e:
            err_msg = 'API Error: {}, url: {}, msg: {}'.format(r.status_code, r.url, r.text)
            self.logger.error(err_msg)
            self.logger.error('Analysis Download output: Failed (id={})'.format(analysis_id))
            sys.exit(1)

    def cancel_generate(self, analysis_id):
        """
        Cancels a currently inputs generation. The analysis status must be `GENERATING_INPUTS`
        """
        try:
            r = self.analyses.generate_cancel(analysis_id)
            self.logger.info('Cancelled Input generation: Id={}'.format(analysis_id))
            return True
        except HTTPError as e:
            err_msg = 'API Error: {}, url: {}:'.format(r.status_code, r.url)
            self.logger.debug(r.text)
            self.logger.debug(err_msg)
            self.logger.info('cancel_generate: Failed')
            return False

    def cancel_analysis(self, analysis_id):
        """
        Cancels a currently running analysis. The analysis must have one of the following
        statuses, `PENDING` or `STARTED`
        """
        try:
            r = self.analyses.run_cancel(analysis_id)
            self.logger.info('Cancelled analysis run: Id={}'.format(analysis_id))
            return True
        except HTTPError as e:
            err_msg = 'API Error: {}, url: {}, msg: {}'.format(r.status_code, r.url, r.text)
            self.logger.error(err_msg)
            self.logger.error('cancel_analysis: Failed')
            return False



#!/usr/bin/env python

import csv
import sys
import msgpack
import struct

"""
Implementation of ktool items conversion tool including
complex item data serialized with msgpack.
"""


def items_to_bin(source, output):
    for row in csv.DictReader(iter(source.readline, '')):
        # item_id,coverage_id,model_data,group_id
        packed_model_data = msgpack.packb(row['model_data'])
        values = (
            int(row['item_id']),
            int(row['coverage_id']),
            int(float(row['group_id'])),
            len(packed_model_data)
        )
        s = struct.Struct('IIII')
        packed_data = s.pack(*values)
        output.write(packed_data)
        output.write(packed_model_data)


def main():

    PY3K = sys.version_info >= (3, 0)

    if PY3K:
        output = sys.stdout.buffer
    else:
        # Python 2 on Windows opens sys.stdin in text mode, and
        # binary data that read from it becomes corrupted on \r\n
        if sys.platform == "win32":
            # set sys.stdin to binary mode
            import os
            import msvcrt
            msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)
        output = sys.stdout

    items_to_bin(sys.stdin, output)


if __name__ == "__main__":

    main()



import io
import os
import random
import re
import string

from collections import Counter

from ..model_preparation.oed import (
    ALLOCATE_TO_ITEMS_BY_PREVIOUS_LEVEL_ALLOC_ID,  # Alloc Rule 2 (Default)
    ALLOCATE_TO_ITEMS_BY_GUL_ALLOC_ID,             # Alloc Rule 1
    NO_ALLOCATION_ALLOC_ID,                        # Alloc Rule 0
)

RUNTYPE_GROUNDUP_LOSS = 'gul'
RUNTYPE_INSURED_LOSS = 'il'
RUNTYPE_REINSURANCE_LOSS = 'ri'

WAIT_PROCESSING_SWITCHES = {
    'full_uncertainty_aep': '-F',
    'wheatsheaf_aep': '-W',
    'sample_mean_aep': '-S',
    'full_uncertainty_oep': '-f',
    'wheatsheaf_oep': '-w',
    'sample_mean_oep': '-s',
    'wheatsheaf_mean_aep': '-M',
    'wheatsheaf_mean_oep': '-m',
}


def print_command(command_file, cmd):
    """
    Writes the supplied command to the end of the generated script

    :param command_file: File to append command to.
    :param cmd: The command to append
    """
    with io.open(command_file, "a", encoding='utf-8') as myfile:
        myfile.writelines(cmd + "\n")


def leccalc_enabled(lec_options):
    """
    Checks if leccalc is enabled in the leccalc options

    :param lec_options: The leccalc options from the analysis settings
    :type lec_options: dict

    :return: True is leccalc is enables, False otherwise.
    """
    if "outputs" not in lec_options:
        return False
    for option in lec_options["outputs"]:
        if lec_options["outputs"][option]:
            return True
    return False


def do_post_wait_processing(runtype, analysis_settings, filename, process_counter):
    if '{}_summaries'.format(runtype) not in analysis_settings:
        return

    for summary in analysis_settings['{}_summaries'.format(runtype)]:
        if "id" in summary:
            summary_set = summary['id']
            if summary.get('aalcalc'):
                cmd = 'aalcalc -K{}_S{}_summaryaalcalc'.format(
                    runtype,
                    summary_set
                )

                process_counter['lpid_monitor_count'] += 1
                cmd = '{} > output/{}_S{}_aalcalc.csv'.format(cmd, runtype, summary_set)
                cmd = '{} & lpid{}=$!'.format(cmd, process_counter['lpid_monitor_count'])
                print_command(filename, cmd)

            if summary.get('lec_output'):
                leccalc = summary.get('leccalc', {})
                if leccalc and leccalc_enabled(leccalc):
                    cmd = 'leccalc {} -K{}_S{}_summaryleccalc'.format(
                        '-r' if leccalc.get('return_period_file') else '',
                        runtype,
                        summary_set
                    )

                    process_counter['lpid_monitor_count'] += 1
                    for option, active in sorted(leccalc['outputs'].items()):
                        if active:
                            switch = WAIT_PROCESSING_SWITCHES.get(option, '')
                            cmd = '{} {} output/{}_S{}_leccalc_{}.csv'.format(cmd, switch, runtype, summary_set,
                                                                              option)

                    cmd = '{} & lpid{}=$!'.format(cmd, process_counter['lpid_monitor_count'])
                    print_command(filename, cmd)


def do_fifos(action, runtype, analysis_settings, process_id, filename, fifo_dir=''):
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    print_command(filename, '{} {}fifo/{}_P{}'.format(action, fifo_dir, runtype, process_id))
    print_command(filename, '')
    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']
            print_command(filename, '{} {}fifo/{}_S{}_summary_P{}'.format(action, fifo_dir, runtype, summary_set, process_id))

            if summary.get('eltcalc'):
                print_command(
                    filename,
                    '{} {}fifo/{}_S{}_summaryeltcalc_P{}'.format(action, fifo_dir, runtype, summary_set, process_id)
                )
                print_command(
                    filename,
                    '{} {}fifo/{}_S{}_eltcalc_P{}'.format(action, fifo_dir, runtype, summary_set, process_id)
                )

            if summary.get('summarycalc'):
                print_command(
                    filename,
                    '{} {}fifo/{}_S{}_summarysummarycalc_P{}'.format(action, fifo_dir, runtype, summary_set, process_id)
                )
                print_command(
                    filename,
                    '{} {}fifo/{}_S{}_summarycalc_P{}'.format(action, fifo_dir, runtype, summary_set, process_id)
                )

            if summary.get('pltcalc'):
                print_command(
                    filename,
                    '{} {}fifo/{}_S{}_summarypltcalc_P{}'.format(action, fifo_dir, runtype, summary_set, process_id)
                )
                print_command(
                    filename,
                    '{} {}fifo/{}_S{}_pltcalc_P{}'.format(action, fifo_dir, runtype, summary_set, process_id)
                )

    print_command(filename, '')


def create_workfolders(runtype, analysis_settings, filename):
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']
            if summary.get('lec_output'):
                if leccalc_enabled(summary['leccalc']):
                    print_command(filename, "mkdir work/{}_S{}_summaryleccalc".format(runtype, summary_set))

            if summary.get('aalcalc'):
                print_command(filename, 'mkdir work/{}_S{}_summaryaalcalc'.format(runtype, summary_set))


def remove_workfolders(runtype, analysis_settings, filename):
    print_command(filename, 'rm -rf work/kat')

    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']
            if summary.get('lec_output'):
                if leccalc_enabled(summary['leccalc']):
                    print_command(filename, 'rm work/{}_S{}_summaryleccalc/*'.format(runtype, summary_set))
                    print_command(filename, 'rmdir work/{}_S{}_summaryleccalc'.format(runtype, summary_set))

            if summary.get('aalcalc'):
                print_command(filename, 'rm -rf work/{}_S{}_summaryaalcalc/*'.format(runtype, summary_set))
                print_command(filename, 'rmdir work/{}_S{}_summaryaalcalc'.format(runtype, summary_set))


def do_make_fifos(runtype, analysis_settings, process_id, filename, fifo_dir=''):
    do_fifos('mkfifo', runtype, analysis_settings, process_id, filename, fifo_dir)


def do_ktools_mem_limit(max_process_id, filename):
    """
        Set each Ktools pipeline to trap and terminate on hitting its
        memory allocation limit

        Limit: Maximum avalible memory / max_process_id
    """
    cmd_mem_limit = 'ulimit -v $(ktgetmem {})'.format(max_process_id)
    print_command(filename, cmd_mem_limit)


def do_remove_fifos(runtype, analysis_settings, process_id, filename, fifo_dir=''):
    do_fifos('rm', runtype, analysis_settings, process_id, filename, fifo_dir)


def do_kats(runtype, analysis_settings, max_process_id, filename, process_counter):
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return False

    anykats = False
    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']

            if summary.get('eltcalc'):
                anykats = True

                cmd = 'kat'
                for process_id in range(1, max_process_id + 1):
                    cmd = '{} work/kat/{}_S{}_eltcalc_P{}'.format(cmd, runtype, summary_set, process_id)

                process_counter['kpid_monitor_count'] += 1
                cmd = '{} > output/{}_S{}_eltcalc.csv & kpid{}=$!'.format(cmd, runtype, summary_set,
                                                                          process_counter['kpid_monitor_count'])
                print_command(filename, cmd)

            if summary.get('pltcalc'):
                anykats = True

                cmd = 'kat'
                for process_id in range(1, max_process_id + 1):
                    cmd = '{} work/kat/{}_S{}_pltcalc_P{}'.format(cmd, runtype, summary_set, process_id)

                process_counter['kpid_monitor_count'] += 1
                cmd = '{} > output/{}_S{}_pltcalc.csv & kpid{}=$!'.format(cmd, runtype, summary_set,
                                                                          process_counter['kpid_monitor_count'])
                print_command(filename, cmd)

            if summary.get("summarycalc"):
                anykats = True

                cmd = 'kat'
                for process_id in range(1, max_process_id + 1):
                    cmd = '{} work/kat/{}_S{}_summarycalc_P{}'.format(cmd, runtype, summary_set, process_id)

                process_counter['kpid_monitor_count'] += 1
                cmd = '{} > output/{}_S{}_summarycalc.csv & kpid{}=$!'.format(cmd, runtype, summary_set,
                                                                              process_counter['kpid_monitor_count'])
                print_command(filename, cmd)

    return anykats


def do_summarycalcs(runtype, analysis_settings, process_id, filename, fifo_dir='',
                    num_reinsurance_iterations=0):

    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    summarycalc_switch = '-f'
    if runtype == RUNTYPE_GROUNDUP_LOSS:
        summarycalc_switch = '-g'

    summarycalc_directory_switch = ""
    if runtype == RUNTYPE_REINSURANCE_LOSS:
        i = num_reinsurance_iterations
        summarycalc_directory_switch = "-p RI_{0}".format(i)

    cmd = 'summarycalc {} {}'.format(summarycalc_switch, summarycalc_directory_switch)
    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']
            cmd = '{0} -{1} {4}fifo/{2}_S{1}_summary_P{3}'.format(cmd, summary_set, runtype, process_id, fifo_dir)

    cmd = '{0} < {1}fifo/{2}_P{3} &'.format(cmd, fifo_dir, runtype, process_id)
    print_command(filename, cmd)


def do_tees(runtype, analysis_settings, process_id, filename, process_counter, fifo_dir=''):
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    for summary in summaries:
        if 'id' in summary:
            process_counter['pid_monitor_count'] += 1
            summary_set = summary['id']
            cmd = 'tee < {}fifo/{}_S{}_summary_P{}'.format(fifo_dir, runtype, summary_set, process_id)

            if summary.get('eltcalc'):
                cmd = '{} {}fifo/{}_S{}_summaryeltcalc_P{}'.format(cmd, fifo_dir, runtype, summary_set, process_id)

            if summary.get('pltcalc'):
                cmd = '{} {}fifo/{}_S{}_summarypltcalc_P{}'.format(cmd, fifo_dir, runtype, summary_set, process_id)

            if summary.get('summarycalc'):
                cmd = '{} {}fifo/{}_S{}_summarysummarycalc_P{}'.format(cmd, fifo_dir, runtype, summary_set, process_id)

            if summary.get('aalcalc'):
                cmd = '{} work/{}_S{}_summaryaalcalc/P{}.bin'.format(cmd, runtype, summary_set, process_id)

            if summary.get('lec_output') and leccalc_enabled(summary['leccalc']):
                cmd = '{} work/{}_S{}_summaryleccalc/P{}.bin'.format(cmd, runtype, summary_set, process_id)

            cmd = '{} > /dev/null & pid{}=$!'.format(cmd, process_counter['pid_monitor_count'])
            print_command(filename, cmd)


def do_any(runtype, analysis_settings, process_id, filename, process_counter, fifo_dir=''):
    summaries = analysis_settings.get('{}_summaries'.format(runtype))
    if not summaries:
        return

    for summary in summaries:
        if 'id' in summary:
            summary_set = summary['id']
            if summary.get('eltcalc'):
                cmd = 'eltcalc -s'
                if process_id == 1:
                    cmd = 'eltcalc'

                process_counter['pid_monitor_count'] += 1
                print_command(
                    filename,
                    "{3} < {5}fifo/{0}_S{1}_summaryeltcalc_P{2} > work/kat/{0}_S{1}_eltcalc_P{2} & pid{4}=$!".format(
                        runtype, summary_set, process_id, cmd, process_counter['pid_monitor_count'], fifo_dir
                    )
                )

            if summary.get("summarycalc"):
                cmd = 'summarycalctocsv -s'
                if process_id == 1:
                    cmd = 'summarycalctocsv'

                process_counter['pid_monitor_count'] += 1
                print_command(
                    filename,
                    '{3} < {5}fifo/{0}_S{1}_summarysummarycalc_P{2} > work/kat/{0}_S{1}_summarycalc_P{2} & pid{4}=$!'.format(
                        runtype, summary_set, process_id, cmd, process_counter['pid_monitor_count'], fifo_dir
                    )
                )

            if summary.get('pltcalc'):
                cmd = 'pltcalc -s'
                if process_id == 1:
                    cmd = 'pltcalc'

                process_counter['pid_monitor_count'] += 1
                print_command(
                    filename,
                    '{3} < {5}fifo/{0}_S{1}_summarypltcalc_P{2} > work/kat/{0}_S{1}_pltcalc_P{2} & pid{4}=$!'.format(
                        runtype, summary_set, process_id, cmd, process_counter['pid_monitor_count'], fifo_dir
                    )
                )

        print_command(filename, '')


def ri(analysis_settings, max_process_id, filename, process_counter, num_reinsurance_iterations, fifo_dir=''):
    for process_id in range(1, max_process_id + 1):
        do_any(RUNTYPE_REINSURANCE_LOSS, analysis_settings, process_id, filename, process_counter, fifo_dir)

    for process_id in range(1, max_process_id + 1):
        do_tees(RUNTYPE_REINSURANCE_LOSS, analysis_settings, process_id, filename, process_counter, fifo_dir)

    for process_id in range(1, max_process_id + 1):
        do_summarycalcs(
            RUNTYPE_REINSURANCE_LOSS, analysis_settings, process_id, filename, fifo_dir, num_reinsurance_iterations)


def il(analysis_settings, max_process_id, filename, process_counter, fifo_dir=''):
    for process_id in range(1, max_process_id + 1):
        do_any(RUNTYPE_INSURED_LOSS, analysis_settings, process_id, filename, process_counter, fifo_dir)

    for process_id in range(1, max_process_id + 1):
        do_tees(RUNTYPE_INSURED_LOSS, analysis_settings, process_id, filename, process_counter, fifo_dir)

    for process_id in range(1, max_process_id + 1):
        do_summarycalcs(RUNTYPE_INSURED_LOSS, analysis_settings, process_id, filename, fifo_dir)


def do_gul(analysis_settings, max_process_id, filename, process_counter, fifo_dir=''):
    for process_id in range(1, max_process_id + 1):
        do_any(RUNTYPE_GROUNDUP_LOSS, analysis_settings, process_id, filename, process_counter, fifo_dir)

    for process_id in range(1, max_process_id + 1):
        do_tees(RUNTYPE_GROUNDUP_LOSS, analysis_settings, process_id, filename, process_counter, fifo_dir)

    for process_id in range(1, max_process_id + 1):
        do_summarycalcs(RUNTYPE_GROUNDUP_LOSS, analysis_settings, process_id, filename, fifo_dir)


def il_make_fifo(analysis_settings, max_process_id, filename, fifo_dir=''):
    for process_id in range(1, max_process_id + 1):
        do_make_fifos(RUNTYPE_INSURED_LOSS, analysis_settings, process_id, filename, fifo_dir)


def do_gul_make_fifo(analysis_settings, max_process_id, filename, fifo_dir=''):
    for process_id in range(1, max_process_id + 1):
        do_make_fifos(RUNTYPE_GROUNDUP_LOSS, analysis_settings, process_id, filename, fifo_dir)


def ri_make_fifo(analysis_settings, max_process_id, filename, fifo_dir=''):
    for process_id in range(1, max_process_id + 1):
        do_make_fifos(RUNTYPE_REINSURANCE_LOSS, analysis_settings, process_id, filename, fifo_dir)


def il_remove_fifo(analysis_settings, max_process_id, filename, fifo_dir=''):
    for process_id in range(1, max_process_id + 1):
        do_remove_fifos(RUNTYPE_INSURED_LOSS, analysis_settings, process_id, filename, fifo_dir)


def do_gul_remove_fifo(analysis_settings, max_process_id, filename, fifo_dir=''):
    for process_id in range(1, max_process_id + 1):
        do_remove_fifos(RUNTYPE_GROUNDUP_LOSS, analysis_settings, process_id, filename, fifo_dir)


def ri_remove_fifo(analysis_settings, max_process_id, filename, fifo_dir=''):
    for process_id in range(1, max_process_id + 1):
        do_remove_fifos(RUNTYPE_REINSURANCE_LOSS, analysis_settings, process_id, filename, fifo_dir)


def do_waits(wait_variable, wait_count, filename):
    """
    Add waits to the script

    :param wait_variable: The type of wait
    :type wait_variable: str

    :param wait_count: The number of processes to wait for
    :type wait_count: int

    :param filename: Script to add waits to
    :type filename: str
    """
    if wait_count > 0:
        cmd = 'wait'
        for pid in range(1, wait_count + 1):
            cmd = '{} ${}{}'.format(cmd, wait_variable, pid)

        print_command(filename, cmd)
        print_command(filename, '')


def do_pwaits(filename, process_counter):
    """
    Add pwaits to the script
    """
    do_waits('pid', process_counter['pid_monitor_count'], filename)


def do_awaits(filename, process_counter):
    """
    Add awaits to the script
    """
    do_waits('apid', process_counter['apid_monitor_count'], filename)


def do_lwaits(filename, process_counter):
    """
    Add lwaits to the script
    """
    do_waits('lpid', process_counter['lpid_monitor_count'], filename)


def do_kwaits(filename, process_counter):
    """
    Add kwaits to the script
    """
    do_waits('kpid', process_counter['kpid_monitor_count'], filename)


def get_getmodel_cmd(
        number_of_samples, gul_threshold, use_random_number_file,
        coverage_output, item_output,
        process_id, max_process_id, **kwargs):
    """
    Gets the getmodel ktools command

    :param number_of_samples: The number of samples to run
    :type number_of_samples: int

    :param gul_threshold: The GUL threshold to use
    :type gul_threshold: float

    :param use_random_number_file: flag to use the random number file
    :type use_random_number_file: bool

    :param coverage_output: The coverage output
    :type coverage_output: str

    :param item_output: The item output
    :type item_output: str

    :return: The generated getmodel command
    """

    cmd = 'eve {0} {1} | getmodel | gulcalc -S{2} -L{3}'.format(
        process_id, max_process_id,
        number_of_samples, gul_threshold)

    if use_random_number_file:
        cmd = '{} -r'.format(cmd)
    if coverage_output != '':
        cmd = '{} -c {}'.format(cmd, coverage_output)
    if item_output != '':
        cmd = '{} -i {}'.format(cmd, item_output)

    return cmd


def genbash(
    max_process_id,
    analysis_settings,
    num_reinsurance_iterations=0,
    fifo_tmp_dir=True,
    mem_limit=False,
    alloc_rule=None,
    bash_trace=False,
    filename='run_kools.sh',
    _get_getmodel_cmd=get_getmodel_cmd,
    custom_args={}
):
    """
    Generates a bash script containing ktools calculation instructions for an
    Oasis model.

    :param max_process_id: The number of processes to create
    :type max_process_id: int

    :param analysis_settings: The analysis settings
    :type analysis_settings: dict

    :param filename: The output file name
    :type filename: string

    :param num_reinsurance_iterations: The number of reinsurance iterations
    :type num_reinsurance_iterations: int

    :param fifo_tmp_dir: When set to True, Create and use FIFO quese in `/tmp/[A-Z,0-9]/fifo`, if False run in './fifo'
    :type fifo_tmp_dir: boolean

    :param alloc_rule: override for the Ktools Allocation rule (1 or 2)
    :type alloc_rule: Int

    :param mem_limit: Flag to set a max memory limit for each ktools process
    :type mem_limit: boolean

    :param get_getmodel_cmd: Method for getting the getmodel command, by default
        ``GenerateLossesCmd.get_getmodel_cmd`` is used.
    :type get_getmodel_cmd: callable
    """
    process_counter = Counter()

    use_random_number_file = False
    gul_output = False
    il_output = False
    ri_output = False
    fifo_queue_dir = ""

    # Alloc Rule input guard - default to '2' if invalid value given
    if alloc_rule not in [ALLOCATE_TO_ITEMS_BY_PREVIOUS_LEVEL_ALLOC_ID,
                          ALLOCATE_TO_ITEMS_BY_GUL_ALLOC_ID,
                          NO_ALLOCATION_ALLOC_ID]:
        alloc_rule = ALLOCATE_TO_ITEMS_BY_PREVIOUS_LEVEL_ALLOC_ID

    # remove the file if it already exists
    if os.path.exists(filename):
        os.remove(filename)

    gul_threshold = analysis_settings.get('gul_threshold', 0)
    number_of_samples = analysis_settings.get('number_of_samples', 0)

    if 'model_settings' in analysis_settings and analysis_settings['model_settings'].get('use_random_number_file'):
        use_random_number_file = True

    if 'gul_output' in analysis_settings:
        gul_output = analysis_settings['gul_output']

    if 'il_output' in analysis_settings:
        il_output = analysis_settings['il_output']

    if 'ri_output' in analysis_settings:
        ri_output = analysis_settings['ri_output']

    print_command(filename, '#!/bin/bash')

    # Use 'set -e' so that any errors in script commands are passed back
    print_command(filename, '')
    print_command(filename, 'set -eux') if bash_trace else print_command(filename, 'set -e')
    print_command(filename, 'set -o pipefail')
    print_command(filename, '')

    print_command(filename, 'rm -R -f output/*')
    if not fifo_tmp_dir:
        print_command(filename, 'rm -R -f fifo/*')
    print_command(filename, 'rm -R -f work/*')
    print_command(filename, '')

    print_command(filename, 'mkdir work/kat')

    # Create tmp dir for FIFO queues (Windows support)
    if fifo_tmp_dir:
        fifo_queue_dir = '/tmp/{}/'.format(
            ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(10))
        )
        print_command(filename, 'mkdir -p {}fifo'.format(fifo_queue_dir))

    if gul_output:
        do_gul_make_fifo(analysis_settings, max_process_id, filename, fifo_queue_dir)
        create_workfolders(RUNTYPE_GROUNDUP_LOSS, analysis_settings, filename)

    if il_output:
        il_make_fifo(analysis_settings, max_process_id, filename, fifo_queue_dir)
        create_workfolders(RUNTYPE_INSURED_LOSS, analysis_settings, filename)

    if ri_output:
        ri_make_fifo(analysis_settings, max_process_id, filename, fifo_queue_dir)
        create_workfolders(RUNTYPE_REINSURANCE_LOSS, analysis_settings, filename)
        print_command(filename, '')

    if ri_output:
        print_command(filename, '')
        print_command(filename, '# --- Do reinsurance loss computes ---')
        print_command(filename, '')
        ri(analysis_settings, max_process_id, filename, process_counter, num_reinsurance_iterations, fifo_queue_dir)

    if il_output:
        print_command(filename, '')
        print_command(filename, '# --- Do insured loss computes ---')
        print_command(filename, '')
        il(analysis_settings, max_process_id, filename, process_counter, fifo_queue_dir)

    if mem_limit:
        print_command(filename, '')
        print_command(filename, '# --- Use Ktools per process memory limit ---')
        print_command(filename, '')
        do_ktools_mem_limit(max_process_id, filename)

    if gul_output:
        print_command(filename, '')
        print_command(filename, '# --- Do ground up loss computes ---')
        print_command(filename, '')
        do_gul(analysis_settings, max_process_id, filename, process_counter, fifo_queue_dir)

    print_command(filename, '')

    for process_id in range(1, max_process_id + 1):

        # ! Should be able to streamline the logic a little
        if num_reinsurance_iterations > 0 and ri_output:

            getmodel_args = {
                'number_of_samples': number_of_samples,
                'gul_threshold': gul_threshold,
                'use_random_number_file': use_random_number_file,
                'coverage_output': '{}fifo/gul_P{}'.format(fifo_queue_dir, process_id),
                'item_output': '-',
                'process_id': process_id,
                'max_process_id': max_process_id
            }
            getmodel_args.update(custom_args)
            getmodel_cmd = _get_getmodel_cmd(**getmodel_args)
            main_cmd = '{2} | fmcalc -a {3} | tee {4}fifo/il_P{0}'.format(
                process_id, max_process_id, getmodel_cmd,
                alloc_rule,
                fifo_queue_dir)

            for i in range(1, num_reinsurance_iterations + 1):
                main_cmd = "{0} | fmcalc -a {3} -n -p RI_{2}".format(
                    main_cmd, os.sep, i, alloc_rule)                    

            main_cmd = "{0} > {1}fifo/ri_P{2} &".format(main_cmd, fifo_queue_dir, process_id)

            print_command(
                filename,
                main_cmd
            )

        elif gul_output and il_output:
            getmodel_args = {
                'number_of_samples': number_of_samples,
                'gul_threshold': gul_threshold,
                'use_random_number_file': use_random_number_file,
                'coverage_output': '{}fifo/gul_P{}'.format(fifo_queue_dir, process_id),
                'item_output': '-',
                'process_id': process_id,
                'max_process_id': max_process_id
            }
            getmodel_args.update(custom_args)
            getmodel_cmd = _get_getmodel_cmd(**getmodel_args)
            main_cmd = '{2} | fmcalc -a {3} > {4}fifo/il_P{0}  &'.format(
                process_id, max_process_id, getmodel_cmd,
                alloc_rule,
                fifo_queue_dir)

            print_command(
                filename,
                main_cmd
            )
        else:
            if gul_output and 'gul_summaries' in analysis_settings:
                getmodel_args = {
                    'number_of_samples': number_of_samples,
                    'gul_threshold': gul_threshold,
                    'use_random_number_file': use_random_number_file,
                    'coverage_output': '-',
                    'item_output': '',
                    'process_id': process_id,
                    'max_process_id': max_process_id
                }
                getmodel_args.update(custom_args)
                getmodel_cmd = _get_getmodel_cmd(**getmodel_args)
                print_command(
                    filename,
                    '{2} > {3}fifo/gul_P{0}  &'.format(process_id, max_process_id, getmodel_cmd, fifo_queue_dir)
                )
            if il_output and 'il_summaries' in analysis_settings:
                getmodel_args = {
                    'number_of_samples': number_of_samples,
                    'gul_threshold': gul_threshold,
                    'use_random_number_file': use_random_number_file,
                    'coverage_output': '',
                    'item_output': '-',
                    'process_id': process_id,
                    'max_process_id': max_process_id
                }
                getmodel_args.update(custom_args)
                getmodel_cmd = _get_getmodel_cmd(**getmodel_args)
                print_command(
                    filename,
                    "{2} | fmcalc -a {3} > {4}fifo/il_P{0}  &".format(
                        process_id, max_process_id, getmodel_cmd,
                        alloc_rule,
                        fifo_queue_dir)
                )

    print_command(filename, '')

    do_pwaits(filename, process_counter)

    if ri_output:
        print_command(filename, '')
        print_command(filename, '# --- Do reinsurance loss kats ---')
        print_command(filename, '')
        do_kats(RUNTYPE_REINSURANCE_LOSS, analysis_settings, max_process_id, filename, process_counter)

    if il_output:
        print_command(filename, '')
        print_command(filename, '# --- Do insured loss kats ---')
        print_command(filename, '')
        do_kats(RUNTYPE_INSURED_LOSS, analysis_settings, max_process_id, filename, process_counter)

    if gul_output:
        print_command(filename, '')
        print_command(filename, '# --- Do ground up loss kats ---')
        print_command(filename, '')
        do_kats(RUNTYPE_GROUNDUP_LOSS, analysis_settings, max_process_id, filename, process_counter)

    do_kwaits(filename, process_counter)

    if mem_limit:
        print_command(filename, '')
        print_command(filename, '# --- Remove per process memory limit ---')
        print_command(filename, '')
        do_ktools_mem_limit(1, filename)

    print_command(filename, '')
    do_post_wait_processing(RUNTYPE_REINSURANCE_LOSS, analysis_settings, filename, process_counter)
    do_post_wait_processing(RUNTYPE_INSURED_LOSS, analysis_settings, filename, process_counter)
    do_post_wait_processing(RUNTYPE_GROUNDUP_LOSS, analysis_settings, filename, process_counter)

    do_awaits(filename, process_counter)  # waits for aalcalc
    do_lwaits(filename, process_counter)  # waits for leccalc

    # Unset '-e' so that a failure to remove temoprary files doesn't kill the entire model run
    print_command(filename, '')
    print_command(filename, 'set +e')
    print_command(filename, '')

    if gul_output:
        do_gul_remove_fifo(analysis_settings, max_process_id, filename, fifo_queue_dir)
        remove_workfolders(RUNTYPE_GROUNDUP_LOSS, analysis_settings, filename)

    print_command(filename, '')

    if ri_output:
        ri_remove_fifo(analysis_settings, max_process_id, filename, fifo_queue_dir)
        remove_workfolders(RUNTYPE_REINSURANCE_LOSS, analysis_settings, filename)

    if il_output:
        il_remove_fifo(analysis_settings, max_process_id, filename, fifo_queue_dir)
        remove_workfolders(RUNTYPE_INSURED_LOSS, analysis_settings, filename)

    # If fifo dir is in /tmp/*/ then clean up
    if re.search(r"((/tmp/)[A-Za-z0-9_-]+(/))", fifo_queue_dir) and fifo_tmp_dir:
        print_command(filename, 'rmdir {}fifo'.format(fifo_queue_dir))
        print_command(filename, 'rmdir {}'.format(fifo_queue_dir))



__all__ = [
    'as_path',
    'empty_dir',
    'PathCleaner',
    'setcwd'
]

import os
import shutil

from contextlib import contextmanager

from .exceptions import OasisException


def as_path(path, label, is_dir=False, preexists=True):
    """
    Processes the path and returns the absolute path.

    If the path does not exist and ``preexists`` is true
    an ``OasisException`` is raised.

    :param path: The path to process
    :type path: str

    :param label: Human-readable label of the path (used for error reporting)
    :type label: str

    :param is_dir: Whether the path is a directory
    :type is_dir: bool

    :param preexists: Flag whether to raise an error if the path
        does not exist.
    :type preexists: bool

    :return: The absolute path of the input path
    """

    if not isinstance(path, str):
        return
    _path = ''.join(path)
    if not os.path.isabs(path):
        _path = os.path.abspath(_path)
    if preexists and not os.path.exists(_path):
        raise OasisException('The path {} ({}) is indicated as preexisting but does not exist'.format(_path, label))
    if is_dir and preexists and not os.path.isdir(_path):
        raise OasisException('The path {} ({}) is indicated as a preexisting directory but is not actually a directory'.format(_path, label))

    return _path


def empty_dir(dir_fp):
    """
    Empties the contents of a directory, but leaves the directory in place.

    :param dir_fp: A pre-existing directory path
    :type dir_fp: str
    """
    _dir_fp = as_path(dir_fp, dir_fp, is_dir=True)

    for p in (os.path.join(_dir_fp, fn) for fn in os.listdir(_dir_fp)):
        os.remove(p) if os.path.isfile(p) else (shutil.rmtree(p) if os.path.isdir(p) else None)


class PathCleaner(object):
    """
    A callable that generates the absolute path of the given path and checks
    that it exists if indicated as preexisting.

    :param label: A user-friendly label for the path (used for error reporting)
    :type label: str

    :param preexists: Flag whether to raise an error if the path
        does not exist.
    :type preexists: bool
    """
    def __init__(self, label, preexists=True):
        self.label = label
        self.preexists = preexists

    def __call__(self, path):
        return as_path(path, self.label, preexists=self.preexists)


@contextmanager
def setcwd(path):
    pwd = os.getcwd()
    os.chdir(str(path))
    yield path
    os.chdir(pwd)



__all__ = [
    'get_fm_terms_oed_columns',
    'get_grouped_fm_profile_by_level',
    'get_grouped_fm_profile_by_level_and_term_group',
    'get_grouped_fm_terms_by_level_and_term_group',
    'get_oed_hierarchy'
]


from collections import OrderedDict
from itertools import groupby

from .defaults import (
    get_default_exposure_profile,
    get_default_accounts_profile,
)
from .fm import (
    SUPPORTED_FM_LEVELS,
    FM_TERMS,
)

def get_grouped_fm_profile_by_level(
    exposure_profile=get_default_exposure_profile(),
    accounts_profile=get_default_accounts_profile()
):
    exp_prof_fm_keys = {k: v for k, v in exposure_profile.items() if 'FMLevel' in v}
    acc_prof_fm_keys = {k: v for k, v in accounts_profile.items() if 'FMLevel' in v}

    comb_prof = {**exp_prof_fm_keys, **acc_prof_fm_keys}

    return OrderedDict({
        int(k): {v['ProfileElementName']: v for v in g}
        for k, g in groupby(sorted(comb_prof.values(), key=lambda v: v['FMLevel']), key=lambda v: v['FMLevel'])
    })


def get_grouped_fm_profile_by_level_and_term_group(
    exposure_profile=get_default_exposure_profile(),
    accounts_profile=get_default_accounts_profile(),
    grouped_profile_by_level=None
):
    grouped = grouped_profile_by_level or get_grouped_fm_profile_by_level(exposure_profile, accounts_profile)

    grouped_fm_term_types = OrderedDict({
        'deductible': FM_TERMS['deductible']['id'],
        'deductiblecode': FM_TERMS['deductible code']['id'],
        'deductibletype': FM_TERMS['deductible type']['id'],
        'deductiblemin': FM_TERMS['min deductible']['id'],
        'deductiblemax': FM_TERMS['max deductible']['id'],
        'limit': FM_TERMS['limit']['id'],
        'limitcode': FM_TERMS['limit code']['id'],
        'limittype': FM_TERMS['limit type']['id'],
        'share': FM_TERMS['share']['id']
    })

    return OrderedDict({
        k: OrderedDict({
            _k: OrderedDict({
                (grouped_fm_term_types.get(v['FMTermType'].lower()) or v['FMTermType'].lower()): v for v in g
            }) for _k, g in groupby(sorted(grouped[k].values(), key=lambda v: v['FMTermGroupID']), key=lambda v: v['FMTermGroupID'])
        }) for k in sorted(grouped)
    })


def get_grouped_fm_terms_by_level_and_term_group(
    exposure_profile=get_default_exposure_profile(),
    accounts_profile=get_default_accounts_profile(),
    grouped_profile_by_level=None,
    grouped_profile_by_level_and_term_group=None,
    lowercase=True
):
    grouped = (
        grouped_profile_by_level_and_term_group or
        get_grouped_fm_profile_by_level_and_term_group(exposure_profile, accounts_profile, grouped_profile_by_level)
    )

    return OrderedDict({
        level_id: OrderedDict({
            tgid: OrderedDict({
                term_type: (
                    (
                        grouped[level_id][tgid][term_type]['ProfileElementName'].lower() if lowercase
                        else grouped[level_id][tgid][term_type]['ProfileElementName']
                    ) if grouped[level_id][tgid].get(term_type) else None
                ) for term_type in [v['id'] for v in FM_TERMS.values()]
            }) for tgid in grouped[level_id]
        }) for level_id in sorted(grouped)
    })


def get_fm_terms_oed_columns(
    fm_terms=get_grouped_fm_terms_by_level_and_term_group(),
    levels=list(SUPPORTED_FM_LEVELS.keys()),
    level_ids=None,
    term_group_ids=[1],
    terms=[v['id'] for v in FM_TERMS.values()],
    remove_nulls=True
):
    level_ids = level_ids or [SUPPORTED_FM_LEVELS[k]['id'] for k in levels]
    _fm_terms = OrderedDict({
        level_id: level_terms_dict
        for level_id, level_terms_dict in fm_terms.items()
        if level_id in level_ids
    })

    cols = [
        _fm_terms[level_id][tgid].get(term)
        for level_id in level_ids
        for tgid in term_group_ids
        for term in terms
    ]

    return cols if not remove_nulls else [col for col in cols if col]


def get_oed_hierarchy(
    exposure_profile=get_default_exposure_profile(),
    accounts_profile=get_default_accounts_profile()
):
    return {v['Key'].lower(): v for k, v in {**exposure_profile, **accounts_profile}.items() if v.get('OEDHierarchy')}



__all__ = [
    'get_calc_rule_ids',
    'get_grouped_fm_profile_by_level_and_term_group',
    'get_grouped_fm_terms_by_level_and_term_group',
    'get_il_input_items',
    'get_layer_ids',
    'get_policytc_ids',
    'write_il_input_files',
    'write_fm_policytc_file',
    'write_fm_profile_file',
    'write_fm_programme_file',
    'write_fm_xref_file'
]

import copy
import os
import sys
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

from itertools import (
    groupby,
)

import pandas as pd
pd.options.mode.chained_assignment = None
import numpy as np

from ..utils.concurrency import (
    multithread,
    Task,
)
from ..utils.calc_rules import get_calc_rules
from ..utils.coverages import SUPPORTED_COVERAGE_TYPES
from ..utils.data import (
    factorize_ndarray,
    fast_zip_arrays,
    get_dataframe,
    merge_dataframes,
    set_dataframe_column_dtypes,
)
from ..utils.defaults import (
    get_default_accounts_profile,
    get_default_exposure_profile,
    get_default_fm_aggregation_profile,
    OASIS_FILES_PREFIXES,
    SOURCE_IDX,
)
from ..utils.exceptions import OasisException
from ..utils.fm import SUPPORTED_FM_LEVELS
from ..utils.log import oasis_log
from ..utils.path import as_path
from ..utils.profiles import (
    get_fm_terms_oed_columns,
    get_grouped_fm_profile_by_level_and_term_group,
    get_grouped_fm_terms_by_level_and_term_group,
    get_oed_hierarchy,
)


def get_layer_ids(accounts_df, accounts_profile=get_default_accounts_profile()):
    """
    Generates a Pandas series of layer IDs given an accounts dataframe - a
    layer ID is an integer index on unique

        ((portfolio num., account num.), policy num.)

    combinations in an account file (or dataframe). The ``PortNumber``,
    ``AccNumber``, ``PolNumber`` columns (or the lowercase equivalents)
    must be present in the accounts dataframe

    :param accounts_df: Accounts dataframe
    :type accounts_df: pandas.DataFrame

    :return: Layer IDs as a Pandas series
    :rtype: pandas.Series
    """
    oed_hierarchy = get_oed_hierarchy(accounts_profile=accounts_profile)
    portfolio_num = oed_hierarchy['portnum']['ProfileElementName'].lower()
    acc_num = oed_hierarchy['accnum']['ProfileElementName'].lower()
    policy_num = oed_hierarchy['polnum']['ProfileElementName'].lower()

    _accounts_df = accounts_df.loc[:, [portfolio_num, acc_num, policy_num]]
    _accounts_df.columns = _accounts_df.columns.str.lower()

    portfolio_nums = _accounts_df[portfolio_num].values
    acc_nums = _accounts_df[acc_num].values
    policy_nums = _accounts_df[policy_num].values

    return np.hstack((
        factorize_ndarray(np.asarray(list(accnum_group)), col_idxs=range(1, 3))[0]
        for _, accnum_group in groupby(fast_zip_arrays(portfolio_nums, acc_nums, policy_nums), key=lambda t: (t[0], t[1]))
    ))


def get_calc_rule_ids(il_inputs_df):
    """
    Returns a Numpy array of calc. rule IDs from a table of IL input items

    :param il_inputs_df: IL input items dataframe
    :type il_inputs_df: pandas.DataFrame

    :return: Numpy array of calc. rule IDs
    :rtype: numpy.ndarray
    """
    calc_rules = get_calc_rules().drop(['desc'], axis=1)
    calc_rules['id_key'] = calc_rules['id_key'].apply(eval)

    terms = ['deductible', 'deductible_min', 'deductible_max', 'limit', 'share', 'attachment']
    terms_indicators = ['{}_gt_0'.format(t) for t in terms]
    types_and_codes = ['ded_type', 'ded_code', 'lim_type', 'lim_code']

    il_inputs_calc_rules_df = il_inputs_df.loc[:, ['item_id'] + terms + terms_indicators + types_and_codes + ['calcrule_id']]
    il_inputs_calc_rules_df.loc[:, terms_indicators] = np.where(il_inputs_calc_rules_df[terms] > 0, 1, 0)
    il_inputs_calc_rules_df['id_key'] = [t for t in fast_zip_arrays(*il_inputs_calc_rules_df.loc[:, terms_indicators + types_and_codes].transpose().values)]
    il_inputs_calc_rules_df = merge_dataframes(il_inputs_calc_rules_df, calc_rules, how='left', on='id_key').fillna(0)
    il_inputs_calc_rules_df['calcrule_id'] = il_inputs_calc_rules_df['calcrule_id'].astype('uint32')

    return il_inputs_calc_rules_df['calcrule_id'].values


def get_policytc_ids(il_inputs_df):
    """
    Returns a Numpy array of policy TC IDs from a table of IL input items

    :param il_inputs_df: IL input items dataframe
    :type il_inputs_df: pandas.DataFrame

    :return: Numpy array of policy TC IDs
    :rtype: numpy.ndarray
    """
    policytc_cols = [
        'layer_id', 'level_id', 'agg_id', 'calcrule_id', 'limit',
        'deductible', 'deductible_min', 'deductible_max', 'attachment',
        'share'
    ]
    fm_policytc_df = il_inputs_df.loc[:, ['item_id'] + policytc_cols].drop_duplicates()
    fm_policytc_df = fm_policytc_df[
        (fm_policytc_df['layer_id'] == 1) |
        (fm_policytc_df['level_id'] == fm_policytc_df['level_id'].max())
    ]

    return factorize_ndarray(fm_policytc_df.loc[:, policytc_cols[3:]].values, col_idxs=range(len(policytc_cols[3:])))[0]


@oasis_log
def get_il_input_items(
    exposure_df,
    gul_inputs_df,
    accounts_df=None,
    accounts_fp=None,
    exposure_profile=get_default_exposure_profile(),
    accounts_profile=get_default_accounts_profile(),
    fm_aggregation_profile=get_default_fm_aggregation_profile()
):
    """
    Generates and returns a Pandas dataframe of IL input items.

    :param exposure_df: Source exposure
    :type exposure_df: pandas.DataFrame

    :param gul_inputs_df: GUL input items
    :type gul_inputs_df: pandas.DataFrame

    :param accounts_df: Source accounts dataframe (optional)
    :param accounts_df: pandas.DataFrame

    :param accounts_fp: Source accounts file path (optional)
    :param accounts_fp: str

    :param exposure_profile: Source exposure profile (optional)
    :type exposure_profile: dict

    :param accounts_profile: Source accounts profile (optional)
    :type accounts_profile: dict

    :param fm_aggregation_profile: FM aggregation profile (optional)
    :param fm_aggregation_profile: dict

    :return: IL inputs dataframe
    :rtype: pandas.DataFrame

    :return Accounts dataframe
    :rtype: pandas.DataFrame
    """
    # Get the grouped exposure + accounts profile - this describes the
    # financial terms found in the source exposure and accounts files,
    # which are for the following FM levels: site coverage (# 1),
    # site pd (# 2), site all (# 3), cond. all (# 6), policy all (# 9),
    # policy layer (# 10).  It also describes the OED hierarchy terms
    # present in the exposure and accounts files, namely portfolio num.,
    # acc. num., loc. num., and cond. num.
    profile = get_grouped_fm_profile_by_level_and_term_group(exposure_profile, accounts_profile)

    if not profile:
        raise OasisException(
            'Unable to get a unified FM profile by level and term group. '
            'Canonical loc. and/or acc. profiles are possibly missing FM term information: '
            'FM term definitions for TIV, deductibles, limit, and/or share.'
        )

    # Get the FM aggregation profile - this describes how the IL input
    # items are to be aggregated in the various FM levels
    fmap = fm_aggregation_profile

    if not fmap:
        raise OasisException(
            'FM aggregation profile is empty - this is required to perform aggregation'
        )

    # Get the OED hierarchy terms profile - this defines the column names for loc.
    # ID, acc. ID, policy no. and portfolio no., as used in the source exposure
    # and accounts files. This is to ensure that the method never makes hard
    # coded references to the corresponding columns in the source files, as
    # that would mean that changes to these column names in the source files
    # may break the method
    oed_hierarchy = get_oed_hierarchy(exposure_profile, accounts_profile)
    loc_num = oed_hierarchy['locnum']['ProfileElementName'].lower()
    acc_num = oed_hierarchy['accnum']['ProfileElementName'].lower()
    policy_num = oed_hierarchy['polnum']['ProfileElementName'].lower()
    portfolio_num = oed_hierarchy['portnum']['ProfileElementName'].lower()
    cond_num = oed_hierarchy['condnum']['ProfileElementName'].lower()

    # Get the FM terms profile (this is a simplfied view of the main grouped
    # profile, containing only information about the financial terms)
    fm_terms = get_grouped_fm_terms_by_level_and_term_group(grouped_profile_by_level_and_term_group=profile)

    # Get the list of financial terms columns for the cond. all (# 6),
    # policy all (# 9) and policy layer (# 10) FM levels - all of these columns
    # are in the accounts file, not the exposure file, so will have to be
    # sourced from the accounts dataframe
    cond_pol_layer_levels = ['cond all', 'policy all', 'policy layer']
    terms_floats = ['deductible', 'deductible_min', 'deductible_max', 'limit', 'attachment', 'share']
    terms_ints = ['ded_code', 'ded_type', 'lim_code', 'lim_type']
    terms = terms_floats + terms_ints
    term_cols_floats = get_fm_terms_oed_columns(
        fm_terms,
        levels=cond_pol_layer_levels,
        terms=terms_floats
    )
    term_cols_ints = get_fm_terms_oed_columns(
        fm_terms,
        levels=cond_pol_layer_levels,
        terms=terms_ints
    )
    term_cols = term_cols_floats + term_cols_ints

    # Set defaults and data types for all the financial terms columns in the
    # accounts dataframe
    defaults = {
        **{t: 0.0 for t in term_cols_floats},
        **{t: 0 for t in term_cols_ints},
        **{cond_num: 0},
        **{portfolio_num: '1'}
    }
    dtypes = {
        **{t: 'str' for t in [acc_num, portfolio_num, policy_num]},
        **{t: 'float64' for t in term_cols_floats},
        **{t: 'uint8' for t in term_cols_ints},
        **{t: 'uint16' for t in [cond_num]},
        **{t: 'uint32' for t in ['layer_id']}
    }

    # Get the accounts frame either directly or from a file path if provided
    accounts_df = accounts_df if accounts_df is not None else get_dataframe(
        src_fp=accounts_fp,
        col_dtypes=dtypes,
        col_defaults=defaults,
        required_cols=(acc_num, policy_num, portfolio_num,),
        empty_data_error_msg='No accounts found in the source accounts (loc.) file',
        memory_map=True,
    )
    accounts_df[SOURCE_IDX['acc']] = accounts_df.index

    if not (accounts_df is not None or accounts_fp):
        raise OasisException('No accounts frame or file path provided')

    # Look for a `layer_id` column in the accounts dataframe - this column
    # will exist if the accounts file has the column - the user has the option
    # of doing this before calling the MDK. The `layer_id` column is simply
    # an enumeration of the unique (portfolio num., acc. num., policy num.)
    # combinations in the accounts file. If the column doesn't exist then
    # a custom method is called that will generate this column and set it
    # in the accounts dataframe
    if 'layer_id' not in accounts_df:
        accounts_df['layer_id'] = get_layer_ids(accounts_df, accounts_profile=accounts_profile)

    # Drop all columns from the accounts dataframe which are not either one of
    # portfolio num., acc. num., policy num., cond. numb., layer ID, or one of
    # the source columns for the financial terms present in the accounts file (the
    # file should contain all financial terms relating to the cond. all (# 6),
    # policy all (# 9) and policy layer (# 10) FM levels)
    usecols = [acc_num, portfolio_num, policy_num, cond_num, 'layer_id', SOURCE_IDX['acc']] + term_cols
    accounts_df.drop([c for c in accounts_df.columns if c not in usecols], axis=1, inplace=True)

    try:
        # Create a list of all the IL columns for the site pd (# 2) and site all (# 3)
        # levels - these columns are in the exposure file, not the accounts
        # file, and so must be sourced from the exposure dataframe
        site_pd_and_site_all_term_cols_floats = get_fm_terms_oed_columns(fm_terms, levels=['site pd', 'site all'], terms=terms_floats)
        site_pd_and_site_all_term_cols_ints = get_fm_terms_oed_columns(fm_terms, levels=['site pd', 'site all'], terms=terms_ints)
        site_pd_and_site_all_term_cols = site_pd_and_site_all_term_cols_floats + site_pd_and_site_all_term_cols_ints

        # Check if any of these columns are missing in the exposure frame, and if so
        # set the missing columns with a default value of 0.0 in the exposure frame
        missing_floats = set(site_pd_and_site_all_term_cols_floats).difference(exposure_df.columns)
        missing_ints = set(site_pd_and_site_all_term_cols_ints).difference(exposure_df.columns)
        defaults = {
            **{t: 0.0 for t in missing_floats},
            **{t: 0 for t in missing_ints}
        }
        if defaults:
            exposure_df = get_dataframe(src_data=exposure_df, col_defaults=defaults)

        # First, merge the exposure and GUL inputs frame to augment the GUL inputs
        # frame with financial terms for level 2 (site PD) and level 3 (site all) -
        # the GUL inputs frame effectively only contains financial terms related to
        # FM level 1 (site coverage)
        gul_inputs_df = merge_dataframes(
            exposure_df.loc[:, site_pd_and_site_all_term_cols + [loc_num]],
            gul_inputs_df,
            join_on=loc_num,
            how='inner'
        )
        gul_inputs_df.rename(columns={'item_id': 'gul_input_id'}, inplace=True)
        dtypes = {t: 'float64' for t in site_pd_and_site_all_term_cols}
        gul_inputs_df = set_dataframe_column_dtypes(gul_inputs_df, dtypes)

        # Construct a basic IL inputs frame by merging the combined exposure +
        # GUL inputs frame above, with the accounts frame, on portfolio no.,
        # account no. and layer ID (by default items in the GUL inputs frame
        # are set with a layer ID of 1)
        il_inputs_df = merge_dataframes(
            gul_inputs_df,
            accounts_df,
            on=[portfolio_num, acc_num, 'layer_id', cond_num],
            how='left',
            drop_duplicates=True
        )

        # Mark the exposure dataframes for deletion
        del exposure_df

        # At this point the IL inputs frame will contain essentially only
        # items for the coverage FM level, but will include multiple items
        # relating to single GUL input items (the higher layer items).

        # If the merge is empty raise an exception - this will happen usually
        # if there are no common acc. numbers between the GUL input items and
        # the accounts listed in the accounts file
        if il_inputs_df.empty:
            raise OasisException(
                'Inner merge of the GUL inputs + exposure file dataframe '
                'and the accounts file dataframe ({}) on acc. number '
                'is empty - '
                'please check that the acc. number columns in the exposure '
                'and accounts files respectively have a non-empty '
                'intersection'.format(accounts_fp)
            )

        # Drop all columns from the IL inputs dataframe which aren't one of
        # necessary columns in the GUL inputs dataframe, or one of policy num.,
        # GUL input item ID, or one of the source columns for the
        # non-coverage FM levels (site PD (# 2), site all (# 3), cond. all (# 6),
        # policy all (# 9), policy layer (# 10))
        usecols = (
            gul_inputs_df.columns.to_list() +
            [policy_num, 'gul_input_id'] +
            ([SOURCE_IDX['loc']] if SOURCE_IDX['loc'] in il_inputs_df else []) +
            ([SOURCE_IDX['acc']] if SOURCE_IDX['acc'] in il_inputs_df else []) +
            site_pd_and_site_all_term_cols +
            term_cols
        )
        il_inputs_df.drop(
            [c for c in il_inputs_df.columns if c not in usecols],
            axis=1,
            inplace=True
        )

        # Mark the GUL inputs frame for deletion - no longer needed
        del gul_inputs_df

        # The coverage FM level (site coverage, # 1) ID
        cov_level_id = SUPPORTED_FM_LEVELS['site coverage']['id']

        # Now set the IL input item IDs, and some other required columns such
        # as the level ID, and initial values for some financial terms,
        # including the calcrule ID and policy TC ID
        il_inputs_df = il_inputs_df.assign(
            level_id=cov_level_id,
            attachment=0.0,
            share=0.0,
            calcrule_id=0,
            policytc_id=0
        )

        # Set data types for the newer columns just added
        dtypes = {
            **{t: 'float64' for t in ['attachment', 'share']},
            **{t: 'uint32' for t in ['level_id', 'calcrule_id', 'policytc_id']}
        }
        il_inputs_df = set_dataframe_column_dtypes(il_inputs_df, dtypes)

        # Drop any items with layer IDs > 1, reset index ad order items by
        # GUL input ID.
        il_inputs_df = il_inputs_df[il_inputs_df['layer_id'] == 1]
        il_inputs_df.reset_index(drop=True, inplace=True)
        il_inputs_df.sort_values('gul_input_id', axis=0, inplace=True)

        # At this stage the IL inputs frame should only contain coverage level
        # layer 1 inputs, and the financial terms are already present from the
        # earlier merge with the exposure and GUL inputs frame - the GUL inputs
        # frame should already contain the coverage level terms

        # The list of financial terms for the sub-layer levels, which are
        # site pd (# 2), site all (# 3), cond. all (# 6), policy all (# 9) -
        # the terms for these levels do not include "attachment" or share",
        # which do exist for the (policy) layer level (# 10); also the
        # layer level terms do not include ded. or limit codes or types
        terms_floats.remove('attachment')
        terms_floats.remove('share')
        terms = terms_floats + terms_ints

        # Steps to filter out any intermediate FM levels which have no
        # financial terms, and also drop all the OED columns for the terms
        # defined for these levels
        def level_has_fm_terms(level, terms):
            try:
                level_terms_cols = get_fm_terms_oed_columns(fm_terms, levels=[level], terms=terms)
                return il_inputs_df.loc[:, level_terms_cols].any().any()
            except KeyError:
                return False

        intermediate_fm_levels = [
            level for level in list(SUPPORTED_FM_LEVELS)[1:-1]
            if level_has_fm_terms(level, terms)
        ]
        fm_levels_with_no_terms = list(set(list(SUPPORTED_FM_LEVELS)[1:-1]).difference(intermediate_fm_levels))
        no_terms_cols = get_fm_terms_oed_columns(fm_terms, levels=fm_levels_with_no_terms, terms=terms)

        il_inputs_df.drop(no_terms_cols, axis=1, inplace=True)

        # Define a list of all supported OED coverage types in the exposure
        supp_cov_types = [v['id'] for v in SUPPORTED_COVERAGE_TYPES.values()]

        # The main loop for processing the financial terms for the sub-layer
        # non-coverage levels - currently these are site pd (# 2), site all (# 3),
        # cond. all (# 6), policy all (# 9).
        #
        # Each level is initially a dataframe copy of the main IL inputs
        # dataframe, which at the start only represents coverage level input
        # items. Using the level terms profile the following steps take place
        # in the loop:
        #
        # (1) financial terms defined for the level are set
        # (2) coverage type filters for the blanket deductibles and limits, if
        # they are defined in the profiles, are applied
        # (3) any blanket deductibles or limits which are expressed as TIV
        # ratios are converted to TIV shares
        #
        # Finally, the processed level dataframe is concatenated with the
        # main IL inputs dataframe, with the financial terms OED columns for
        # level removed
        for level in intermediate_fm_levels:
            level_id = SUPPORTED_FM_LEVELS[level]['id']
            level_terms = [t for t in terms if fm_terms[level_id][1].get(t)]
            level_term_cols = get_fm_terms_oed_columns(fm_terms, level_ids=[level_id], terms=terms)
            level_df = il_inputs_df[il_inputs_df['level_id'] == cov_level_id].drop_duplicates()
            level_df['level_id'] = level_id

            agg_key = [v['field'].lower() for v in fmap[level_id]['FMAggKey'].values()]
            level_df['agg_id'] = factorize_ndarray(level_df.loc[:, agg_key].values, col_idxs=range(len(agg_key)))[0]

            if level == 'cond all':
                level_df.loc[:, level_term_cols] = level_df.loc[:, level_term_cols].fillna(0)
            else:
                level_df.loc[:, level_term_cols] = level_df.loc[:, level_term_cols].fillna(method='ffill')
                level_df.loc[:, level_term_cols] = level_df.loc[:, level_term_cols].fillna(0)

            level_df.loc[:, level_terms] = level_df.loc[:, level_term_cols].values

            level_df['deductible'] = np.where(
                level_df['coverage_type_id'].isin((profile[level_id][1].get('deductible') or {}).get('CoverageTypeID') or supp_cov_types),
                level_df['deductible'],
                0
            )
            level_df['limit'] = np.where(
                level_df['coverage_type_id'].isin((profile[level_id][1].get('limit') or {}).get('CoverageTypeID') or supp_cov_types),
                level_df['limit'],
                0
            )

            il_inputs_df = pd.concat([il_inputs_df, level_df], sort=True, ignore_index=True)
            il_inputs_df.drop(level_term_cols, axis=1, inplace=True)

        # Resequence the item IDs, as the earlier repeated concatenation of
        # the intermediate level frames may have produced a non-sequential index
        il_inputs_df['item_id'] = il_inputs_df.index + 1

        # Process the layer FM level (policy layer, # 10) inputs separately - we
        # start with merging the coverage level layer 1 items with the accounts
        # dataframe to create a separate layer level frame, on which further
        # processing is done
        cov_level_layer1_df = il_inputs_df[il_inputs_df['level_id'] == cov_level_id]
        layer_df = merge_dataframes(
            cov_level_layer1_df,
            accounts_df,
            on=[portfolio_num, acc_num],
            how='inner'
        )

        # Remove the source columns for all non-layer FM levels - this includes the
        # site pd (# 2), site all (# 3), cond. all (# 6), policy all (# 9) FM levels
        cond_all_and_pol_all_term_cols = get_fm_terms_oed_columns(fm_terms, levels=['cond all', 'policy all'])
        layer_df.drop(
            [c for c in layer_df.columns if c in site_pd_and_site_all_term_cols + cond_all_and_pol_all_term_cols],
            axis=1, inplace=True
        )

        # The layer FM level (policy layer, # 10) ID
        layer_level_id = SUPPORTED_FM_LEVELS['policy layer']['id']

        # Set the layer level, layer IDs and agg. IDs
        layer_df['level_id'] = layer_level_id
        agg_key = [v['field'].lower() for v in fmap[layer_level_id]['FMAggKey'].values()]
        layer_df['agg_id'] = factorize_ndarray(layer_df.loc[:, agg_key].values, col_idxs=range(len(agg_key)))[0]

        # The layer level financial terms
        terms = ['limit', 'attachment', 'share']

        # Process the financial terms for the layer level
        term_cols = get_fm_terms_oed_columns(fm_terms, levels=['policy layer'], terms=terms)
        layer_df.loc[:, term_cols] = layer_df.loc[:, term_cols].where(layer_df.notnull(), 0.0).values
        layer_df.loc[:, terms] = layer_df.loc[:, term_cols].values
        layer_df['limit'] = layer_df['limit'].where(layer_df['limit'] != 0, 9999999999)
        layer_df['share'] = layer_df['share'].where(layer_df['share'] != 0, 1.0)
        layer_df['deductible'] = 0
        layer_df.loc[:, ['ded_code', 'ded_type', 'lim_code', 'lim_type']] = 0

        # Join the IL inputs and layer level frames, drop the FM terms
        # source columns for the layer level, and mark the layer level dataframe
        # for deletion
        il_inputs_df = pd.concat([il_inputs_df, layer_df], sort=True, ignore_index=True)
        il_inputs_df.drop(term_cols, axis=1, inplace=True)
        del layer_df

        # Resequence the level IDs and item IDs, but also store the "original"
        # FM level IDs (before the resequencing)
        il_inputs_df['orig_level_id'] = il_inputs_df['level_id']
        il_inputs_df['level_id'] = factorize_ndarray(il_inputs_df.loc[:, ['level_id']].values, col_idxs=[0])[0]
        il_inputs_df['item_id'] = il_inputs_df.index + 1

        # Set datatypes again for the deductible code and type columns, as
        # they may have changed since the processing of the intermediate level
        # terms
        dtypes = {t: 'uint8' for t in ['ded_code', 'ded_type', 'lim_code', 'lim_type']}
        il_inputs_df = set_dataframe_column_dtypes(il_inputs_df, dtypes)

        # Set the calc. rule IDs
        il_inputs_df['calcrule_id'] = get_calc_rule_ids(il_inputs_df)

        # Set the policy TC IDs
        il_inputs_df['policytc_id'] = get_policytc_ids(il_inputs_df)

        # Final setting of data types before returning the IL input items
        dtypes = {
            **{t: 'uint32' for t in ['agg_id', 'item_id', 'layer_id', 'level_id', 'orig_level_id', 'calcrule_id', 'policytc_id']},
            **{t: 'uint16' for t in [cond_num]},
            **{t: 'uint8' for t in ['ded_code', 'ded_type', 'lim_code', 'lim_type']}
        }
        il_inputs_df = set_dataframe_column_dtypes(il_inputs_df, dtypes)

    except (AttributeError, KeyError, IndexError, TypeError, ValueError) as e:
        raise OasisException from e

    return il_inputs_df, accounts_df


@oasis_log
def write_fm_policytc_file(il_inputs_df, fm_policytc_fp, chunksize=100000):
    """
    Writes an FM policy T & C file.

    :param il_inputs_df: IL inputs dataframe
    :type il_inputs_df: pandas.DataFrame

    :param fm_policytc_fp: FM policy TC file path
    :type fm_policytc_fp: str

    :return: FM policy TC file path
    :rtype: str
    """
    try:
        fm_policytc_df = il_inputs_df.loc[:, ['layer_id', 'level_id', 'agg_id', 'policytc_id']]
        fm_policytc_df.drop_duplicates().to_csv(
            path_or_buf=fm_policytc_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(fm_policytc_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException from e

    return fm_policytc_fp


@oasis_log
def write_fm_profile_file(il_inputs_df, fm_profile_fp, chunksize=100000):
    """
    Writes an FM profile file.

    :param il_inputs_df: IL inputs dataframe
    :type il_inputs_df: pandas.DataFrame

    :param fm_profile_fp: FM profile file path
    :type fm_profile_fp: str

    :return: FM profile file path
    :rtype: str
    """
    try:
        cols = ['policytc_id', 'calcrule_id', 'deductible', 'deductible_min', 'deductible_max', 'attachment', 'limit', 'share']
        fm_profile_df = il_inputs_df.loc[:, cols]

        fm_profile_df.loc[:, cols[2:]] = fm_profile_df.loc[:, cols[2:]].round(7).values

        fm_profile_df.rename(
            columns={
                'deductible': 'deductible1',
                'deductible_min': 'deductible2',
                'deductible_max': 'deductible3',
                'attachment': 'attachment1',
                'limit': 'limit1',
                'share': 'share1'
            },
            inplace=True
        )
        fm_profile_df = fm_profile_df.drop_duplicates()

        fm_profile_df = fm_profile_df.assign(share2=0.0, share3=0.0)

        cols = ['policytc_id', 'calcrule_id', 'deductible1', 'deductible2', 'deductible3', 'attachment1', 'limit1', 'share1', 'share2', 'share3']
        fm_profile_df.loc[:, cols].to_csv(
            path_or_buf=fm_profile_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(fm_profile_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException from e

    return fm_profile_fp


@oasis_log
def write_fm_programme_file(il_inputs_df, fm_programme_fp, chunksize=100000):
    """
    Writes an FM programme file.

    :param il_inputs_df: IL inputs dataframe
    :type il_inputs_df: pandas.DataFrame

    :param fm_programme_fp: FM programme file path
    :type fm_programme_fp: str

    :return: FM programme file path
    :rtype: str
    """
    try:
        fm_programme_df = pd.concat(
            [
                il_inputs_df[il_inputs_df['level_id'] == il_inputs_df['level_id'].min()].loc[:, ['agg_id']].assign(level_id=0),
                il_inputs_df.loc[:, ['level_id', 'agg_id']]
            ]
        ).reset_index(drop=True)

        min_level, max_level = 0, fm_programme_df['level_id'].max()

        fm_programme_df = pd.DataFrame(
            {
                'from_agg_id': fm_programme_df[fm_programme_df['level_id'] < max_level]['agg_id'],
                'level_id': fm_programme_df[fm_programme_df['level_id'] > min_level]['level_id'].reset_index(drop=True),
                'to_agg_id': fm_programme_df[fm_programme_df['level_id'] > min_level]['agg_id'].reset_index(drop=True)
            }
        ).dropna(axis=0).drop_duplicates()

        dtypes = {t: 'uint32' for t in fm_programme_df.columns}
        fm_programme_df = set_dataframe_column_dtypes(fm_programme_df, dtypes)

        fm_programme_df.to_csv(
            path_or_buf=fm_programme_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(fm_programme_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException from e

    return fm_programme_fp


@oasis_log
def write_fm_xref_file(il_inputs_df, fm_xref_fp, chunksize=100000):
    """
    Writes an FM xref file.

    :param il_inputs_df: IL inputs dataframe
    :type il_inputs_df: pandas.DataFrame

    :param fm_xref_fp: FM xref file path
    :type fm_xref_fp: str

    :return: FM xref file path
    :rtype: str
    """
    try:
        cov_level_layers_df = il_inputs_df[il_inputs_df['level_id'] == il_inputs_df['level_id'].max()]
        pd.DataFrame(
            {
                'output': factorize_ndarray(cov_level_layers_df.loc[:, ['gul_input_id', 'layer_id']].values, col_idxs=range(2))[0],
                'agg_id': cov_level_layers_df['gul_input_id'],
                'layer_id': cov_level_layers_df['layer_id']
            }
        ).drop_duplicates().to_csv(
            path_or_buf=fm_xref_fp,
            encoding='utf-8',
            mode=('w' if os.path.exists(fm_xref_fp) else 'a'),
            chunksize=chunksize,
            index=False
        )
    except (IOError, OSError) as e:
        raise OasisException from e

    return fm_xref_fp


@oasis_log
def write_il_input_files(
    il_inputs_df,
    target_dir,
    oasis_files_prefixes=copy.deepcopy(OASIS_FILES_PREFIXES['il'])
):
    """
    Writes standard Oasis IL input files to a target directory using a
    pre-generated dataframe of IL inputs dataframe. The files written are
    ::

        fm_policytc.csv
        fm_profile.csv
        fm_programme.csv
        fm_xref.csv

    :param il_inputs_df: IL inputs dataframe
    :type exposure_df: pandas.DataFrame

    :param target_dir: Target directory in which to write the files
    :type target_dir: str

    :param oasis_files_prefixes: Oasis IL input file name prefixes
    :param oasis_files_prefixes: dict

    :return: IL input files dict
    :rtype: dict
    """
    # Clean the target directory path
    target_dir = as_path(target_dir, 'Target IL input files directory', is_dir=True, preexists=False)

    # Set chunk size for writing the CSV files - default is the minimum of 100K
    # or the IL inputs frame size
    chunksize = min(2 * 10**5, len(il_inputs_df))

    # A dict of IL input file names and file paths
    il_input_files = {
        fn: os.path.join(target_dir, '{}.csv'.format(oasis_files_prefixes[fn])) for fn in oasis_files_prefixes
    }

    # IL input file writers have the same filename prefixes as the input files
    # and we use this property to dynamically retrieve the methods from this
    # module
    this_module = sys.modules[__name__]

    # Create a generator of ``oasislmf.utils.concurrency.Task`` objects to
    # represent the writing of the individual input files, one per file
    tasks = (
        Task(
            getattr(this_module, 'write_{}_file'.format(fn)),
            args=(il_inputs_df.copy(deep=True), il_input_files[fn], chunksize,),
            key=fn
        )
        for fn in il_input_files
    )

    for _, _ in multithread(tasks, pool_size=len(il_input_files)):
        pass

    return il_input_files
